/*
 * Test the eepoll (polling Linux AIO inside epoll) patch
 * It does I/O using both a disk fd and a pipe
 *
 * It's not using any new function of Flex, so it should just
 * work with Linux AIO + epoll patch + eepoll patch
 */

#include <stdio.h>
#include <flex.h>
#include <epoll.h>   /* FIXME: This should go into flex.h */
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/poll.h>

int epfd;
io_context_t ioctx;

inline void assert(const int t, const char *msg) {
  if (!t) {
    perror(msg);
    exit(-1);
  }
}

int main(int argc, char *argv[]) {
  
  struct epoll_event evtdisk, evtpipe, epevents[2];
  int pipefds[2];
  int fd, n;
  struct iocb iocb, *piocb = &iocb;
  char buf[1024];
  struct timespec ts = {.tv_sec = 0, .tv_nsec = 0};
  struct io_event ioevt;

  if ((epfd = epoll_create (1024)) < 0) {
    perror("Cannot initialize epoll\n");
    exit(-1);
  }

  if (io_setup (1024, &ioctx) < 0) {
    perror("Cannot initialize linux aio\n");
    exit(-1);
  }

  /* Add the AIO context to epoll */
  evtdisk.events = 0;
  if (epoll_ctl (epfd, EPOLL_CTL_ADD | 0x2000, (int) ioctx, &evtdisk) < 0) {
    perror("Cannot add AIO context into epoll.  "
	   "Do you have the eepoll patch installed?");
    exit(-1);
  }
  
  /* Create the pipe and add it to epoll */
  if (pipe (pipefds) < 0) {
    perror("Cannot create pipes");
    exit(-1);
  }
  evtpipe.events = POLLIN;
  if (epoll_ctl (epfd, EPOLL_CTL_ADD, pipefds[0], &evtpipe) < 0) {
    perror ("Cannot add pipe fd to epoll");
    exit(-1);
  }

  if ((fd = open ("/etc/services", O_RDONLY)) < 0) {
    perror("Cannot open test file\n");
    exit(-1);
  }

  /* Issue a disk read */
  io_prep_pread(&iocb, fd, buf, 1024, 0);
  assert (io_submit(ioctx, 1, &piocb) == 1,
	  "Cannot submit request");

  /* Wait for epoll events */
  n = epoll_wait(epfd, epevents, 2, -1);
  assert (n > 0, "Cannot get epoll events");
  assert (n == 1, "Get more than 1 epoll events");

  /* It should be an AIO ready event */
  assert (epevents[0].events & 0x2000, "Got an epoll event other than an AIO ready event");

  /* Go get the AIO event */
  n = io_getevents(ioctx, 1, 1, &ioevt, &ts);
  assert (n > 0, "Cannot get Linux AIO event");
  assert (ioevt.obj == &iocb, "IOCB of AIO event does not match");

  printf ("Done successfully\n");
  return 0;
}
