/* pipetest.c
 *	Scalability test of async poll functionality on pipes.
 *	Copyright 2002 Red Hat, Inc.
 *	Portions Copyright 2001 Davide Libenzi <davidel@xmailserver.org>.
 *	<include GPL>
 *
 *      Feng Zhou (epoll-over-aio test)
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/mman.h>

#include <asm/unistd.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>


#define PAGE_SIZE 4096

#include <linux/ioctl.h>
#include <linux/linkage.h>
#include <error.h>
#include <sys/epoll.h>
#include <libaio.h>

//#define DEBUG

#if !defined(DEBUG)
#define dprintf(x...)	do {} while(0)
#define d2printf(x...)	do {} while(0)
#else
#define dprintf(x...)	printf(x)
#define d2printf(x...)	printf(x)
#endif

enum {
	MODE_POLL,
	MODE_SYS_EPOLL,
	MODE_EPIO,
} mode = MODE_POLL;

const char *modes[] = {
	"poll",
	"sys-epoll",
	"epio",
};

int poll_may_complete;
io_context_t ctx;

int epoll_fd = -1;
char *epoll_map;

int done;

#define FD_SLOP	10
#define MAX_FDS 1024	

#define READ	0
#define WRITE	1

struct {
	int fds[2];
} pipefds[MAX_FDS/2];

int nr_pipes;

long	nr_token_passes;
long	max_generation;
int	max_threads = 1;
int	threads_complete;

struct token {
	long	generation;
	int	tofd;
	int	thread;
};

//#define BUFSIZE		sizeof(struct token)
int BUFSIZE = sizeof(struct token);

struct fdinfo {
	struct iocb	iocb;
	int		zero;
	//struct iocb	iocb2;
	int		one;
	int		fd;
	int		pipe_idx;
	int		active:1;
	int		rw:1;
	char		*buf;
	//char		buf2[BUFSIZE];
} fdinfo[MAX_FDS];

#define queue_iocb(iocb)	(pending_iocbs[nr_pending_iocbs++] = (iocb))

struct pollfd	pollfds[MAX_FDS];
int nr_pollfds;

struct iocb *pending_iocbs[MAX_FDS*3];
int nr_pending_iocbs;

struct token pending_tokens[MAX_FDS];
int nr_pending_tokens;

void pexit(char *msg)
{
	perror(msg);
	exit(1);
}

void really_send_toke(struct token *toke)
{
	struct fdinfo *inf;
	int *fds;
	int fd;
        int res;
	int pipe_idx;// = nr_pipes - toke->thread - 1;

	/* destination pipe */
	pipe_idx = (toke->thread + toke->generation * 17) % nr_pipes;
	/* pipe_idx = rand () % nr_pipes; */

	fds = pipefds[pipe_idx].fds;
	fd = fds[WRITE];
	inf = &fdinfo[fd];
	toke->tofd = fds[READ];

	res = write(fds[WRITE], toke, BUFSIZE);
	d2printf("write: generation=%ld, dest_pipe=%u, tofd=%d, thread=%d\n", 
		 toke->generation,
		 pipe_idx,
		 toke->tofd, toke->thread);
	if (res != BUFSIZE) {
	  printf("write = %d (%s)", res, strerror(errno));
	  exit(1);
	}
}

void send_pending_tokes(void)
{
	int i;
	for (i=0; i<nr_pending_tokens; i++)
		really_send_toke(&pending_tokens[i]);
	nr_pending_tokens = 0;
}

void send_toke(struct token *toke)
{
	pending_tokens[nr_pending_tokens++] = *toke;
}

void process_token(int fd, struct token *toke, int nr)
{
	if (nr != BUFSIZE)
		fprintf(stderr, "process_token: nr == %d (vs %d)\n",
			nr, BUFSIZE);
	assert(nr == BUFSIZE);
	assert(toke->tofd == fd);
	nr_token_passes++;
	toke->generation++;

	dprintf("passed %ld\n", nr_token_passes);

	if (toke->generation < max_generation)
		send_toke(toke);
	else {
		dprintf("thread %d complete, %ld passes\n", toke->thread, nr_token_passes);
		threads_complete++;
		if (threads_complete >= max_threads)
			done = 1;
	}
}

void read_and_process_token(int fd)
{
	//char buf[BUFSIZE];
	struct token *toke;
	struct fdinfo *inf = &fdinfo[fd];
	char *buf = inf->buf;
	int nr;
	nr = read(fd, buf, BUFSIZE);
	if (-1 == nr)
		pexit("read");
	/* kludge: works around epoll edge notification bug */
	toke = (struct token *)buf;
	dprintf("Received token\n");
	while (nr >= BUFSIZE) {
		process_token(fd, toke, BUFSIZE);
		toke++;
		nr -= BUFSIZE;
	}
	assert(nr == 0);
}

void makeapipe(int idx)
{
	int *fds = pipefds[idx].fds;
	struct fdinfo *inf;
	int i;

	if (pipe(fds))
		pexit("pipe");

	for (i=0; i<2; i++) {
		int fl = fcntl(fds[i], F_GETFL);
		fl |= O_NONBLOCK;
		fcntl(fds[i], F_SETFL, fl);
	}

	inf = &fdinfo[fds[READ]];
	inf->buf = calloc(1, BUFSIZE);
	assert(inf->buf != NULL);
	assert(inf->active == 0);
	inf->active = 1;
	inf->rw = READ;
	inf->fd = fds[READ];
	inf->pipe_idx = idx;

        if (mode == MODE_POLL) {
                pollfds[nr_pollfds].fd = fds[READ];
                pollfds[nr_pollfds].events = POLLIN;
                pollfds[nr_pollfds].revents = 0;
                nr_pollfds++;
        }

	if (mode == MODE_SYS_EPOLL || mode == MODE_EPIO) {
		struct epoll_event evt;
		evt.events = POLLIN;
		evt.data.fd = fds[READ];
		epoll_ctl(epoll_fd, EPOLL_CTL_ADD, fds[READ], &evt);
	}

	inf = &fdinfo[fds[WRITE]];
	assert(inf->active == 0);
	inf->active = 1;
	inf->rw = WRITE;
	inf->fd = fds[WRITE];
	inf->pipe_idx = idx;
}

void makepipes(int nr)
{
	int i;
	for (i=0; i<nr; i++)
		makeapipe(i);
	nr_pipes = nr;
}

void poll_main_loop(void)
{
	int res;
	while (!done) {
		int i, j;

		dprintf("poll_main_loop: poll(%d)\n", nr_pollfds);
		send_pending_tokes();
		res = poll(pollfds, nr_pollfds, -1);
		if (res <= 0)
			pexit("poll");

		dprintf("got %d\n", res);

		for (i=0,j=0; i<res; j++) {
			if (pollfds[j].revents & POLLIN) {
				read_and_process_token(pollfds[j].fd);
				i++;
			}
		}		
	}
}

void sys_epoll_setup(void)
{
	epoll_fd = epoll_create(MAX_FDS);
	if (-1 == epoll_fd)
		pexit("epoll_create");
}
struct epoll_event evts[100];

/* This function is stolen from dphttpd.c */
void sys_epoll_main_loop(void)
{
	int ii, nfds;

	while (!done) {
		send_pending_tokes();

		nfds = epoll_wait(epoll_fd, evts, 100, -1);
		if (nfds < 0)
			pexit("epoll_wait");
		for (ii = 0; ii < nfds; ii++) {
			if (evts[ii].events & POLLIN)
				read_and_process_token(evts[ii].data.fd);
		}
	}
}

void epio_setup(void) {
  int res;
  /* create epoll fd */
  epoll_fd = epoll_create(1024);
  if (epoll_fd < 0) {
    fprintf(stderr, "epio_setup: cannot create epoll fd %d (%s)\n", epoll_fd, 
	   strerror(-epoll_fd));
    exit(-1);
  }
  /* create the aio context */
  res = io_setup(1024, &ctx);
  if (res < 0) {
    fprintf(stderr, "epio_setup: cannot create aio context %d (%s)\n", res, 
	   strerror(-res));
    exit(-1);
  }
#if 0
    res = io_modify(ctx, IO_SET_EPOLLFD, (void *)epoll_fd);

    struct iocb req, *preq = &req;
    memset(&req, 0, sizeof(struct iocb));
    req.aio_lio_opcode = IO_CMD_SET_EPOLLFD;
    req.aio_fildes = epoll_fd;
    res = io_submit(ctx, 1, &preq);

#else
    res = io_set_epollfd(ctx, epoll_fd);
#endif

  if (res < 0) {
    fprintf(stderr, "epio_setup: %d (%s)\n", res, strerror(-res));
    exit(1);
  }
}

void epio_main_loop(void)
{
  struct io_event events[100];
  int i, res;

  while (!done) {
    send_pending_tokes();

    res = io_getevents(ctx, 1, MAX_FDS, events, 0);
    if (res < 0) {
      dprintf("io_getevents: %d (%s)", res, strerror(-res));
      exit(1);
    }
    for (i=0; i<res; i++) {
      int revents = events[i].res;
      int fd = events[i].res2;
      if (events[i].obj != IOCB_POLL) {
	fprintf(stderr, "Got a non-epoll event!\n");
	continue;
      }
      dprintf("Event: fd=%d, revents=0x%x\n", fd, revents);
      if (revents & POLLIN) {
	read_and_process_token(fd);
      }
    }
  }
}

void seedthreads(int nr)
{
	struct token toke;
	int i;

	for (i=0; i<nr; i++) {
		toke.generation = 0;
		toke.thread = i;
		toke.tofd = -1;
		send_toke(&toke);
	}
}

int main(int argc, char *argv[])
{
	struct timeval stv, etv;
	int nr;
	long long usecs, passes_per_sec;

	while (argc > 1) {
		if (0 == strcmp(argv[1], "--epio")) {
			mode = MODE_EPIO;
		} else if (0 == strcmp(argv[1], "--epoll")) {
			mode = MODE_SYS_EPOLL;
		} else if (0 == strcmp(argv[1], "--poll")) {
			mode = MODE_POLL;
		} else if (0 == strcmp(argv[1], "--sys-epoll")) {
			mode = MODE_SYS_EPOLL;
		} else if (0 == strcmp(argv[1], "--bufsize")) {
			argv++,argc--;
			BUFSIZE = atoi(argv[1]);
			assert(BUFSIZE > sizeof(struct token));
		} else
			break;
		argv++,argc--;
	}

	if (argc != 4) {
		fprintf(stderr, "usage: pipetest [--epio| --poll | --epoll] [--bufsize] <num pipes> <message threads> <max generation>\n");
		fprintf(stderr, "\tuse --epio for testing epoll-over-aio event queue\n");
		return 2;
	}

	nr = atoi(argv[1]);
	max_threads = atoi(argv[2]);
	max_generation = atol(argv[3]);
	fprintf(stderr, "using %d pipe pairs, %d message threads, %ld generations, %d bufsize\n",
		nr, max_threads, max_generation, BUFSIZE);

	if (nr < 2) {
		printf("uhm, please specify at least 2 pipe pairs\n");
		exit(1);
	}
	if (nr >= (MAX_FDS/2 - FD_SLOP)) {
		printf("%d exceeds limit of %d pipe pairs.\n", nr,
			(MAX_FDS/2-FD_SLOP));
		exit(1);
	}

	if (mode == MODE_EPIO) {
                epio_setup();
	}
	if (mode == MODE_SYS_EPOLL)
		sys_epoll_setup();

	makepipes(nr);

	send_pending_tokes();

	gettimeofday(&stv, NULL);

	seedthreads(max_threads);

	if (mode == MODE_POLL)
		poll_main_loop();

	if (mode == MODE_SYS_EPOLL)
		sys_epoll_main_loop();

	if (mode == MODE_EPIO) {
	        epio_main_loop();
	}

	gettimeofday(&etv, NULL);

	etv.tv_sec -= stv.tv_sec;
	etv.tv_usec -= stv.tv_usec;
	if (etv.tv_usec < 0) {
		etv.tv_usec += 1000000;
		etv.tv_sec -= 1;
	}

	fprintf(stderr, "Ok! Mode %s: %ld passes in %ld.%06ld seconds\n",
		modes[mode],
		nr_token_passes,
		etv.tv_sec, etv.tv_usec);

	
	usecs = etv.tv_usec + etv.tv_sec * 1000000LL;
	passes_per_sec = nr_token_passes * 1000000LL * 100;
	passes_per_sec /= usecs;

	fprintf(stderr, "passes_per_sec: %Ld.%02Ld\n",
		passes_per_sec / 100,
		passes_per_sec % 100);

	printf ("%d %Ld.%02Ld\n", nr, 
		passes_per_sec / 100,
		passes_per_sec % 100);

	return 0;
}


